from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict

from ..utils import snake_case_keys

if TYPE_CHECKING:
    from ..reddit import Reddit


class aPRAWBase:

    def __init__(self, reddit: 'Reddit', data: Dict[str, Any]):
        self.reddit = reddit
        self._update(data)

    def _update(self, data: Dict[str, Any]):
        self.data = data

        d = snake_case_keys(data)
        for key in d:
            if not hasattr(self, key):
                setattr(self, key, d[key])

        if "created_utc" in data:
            self.created_utc = datetime.utcfromtimestamp(data["created_utc"])



from .comment import Comment
from .redditor import Redditor
from .submission import Submission
from .subreddit import Subreddit
from .listing_generator import ListingGenerator
from .user import User
from .subreddit_wiki import WikipageRevision



from typing import TYPE_CHECKING, Dict

from ..endpoints import API_PATH
from ..utils import snake_case_keys
from .apraw_base import aPRAWBase

if TYPE_CHECKING:
    from .subreddit import Subreddit
    from .redditor import Redditor
    from ..reddit import Reddit


class SubredditModmail:

    def __init__(self, subreddit: 'Subreddit'):
        self.subreddit = subreddit

    async def conversations(self) -> 'ModmailConversation':
        req = await self.subreddit.reddit.get_request(API_PATH["modmail_conversations"], entity=self.subreddit.display_name)
        for id in req["conversations"]:
            yield ModmailConversation(self.subreddit.reddit, req["conversations"][id])


class ModmailConversation(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict,
                 owner: 'Subreddit' = None):
        super().__init__(reddit, data)

        self._data = None

        self.id = data["id"]
        self._owner = owner

    async def owner(self) -> 'Subreddit':
        if self._owner is None:
            self._owner = await self.reddit.subreddit(self.data["owner"]["displayName"])
        return self._owner

    async def messages(self) -> 'ModmailMessage':
        full_data = await self.full_data()
        for msg_id in full_data["messages"]:
            yield ModmailMessage(self, full_data["messages"][msg_id])

    async def full_data(self) -> Dict:
        if self._data is None:
            self._data = await self.reddit.get_request(API_PATH["modmail_conversation"].format(id=self.id))
        return self._data


class ModmailMessage:

    def __init__(self, conversation: ModmailConversation, data: Dict):
        self.conversation = conversation
        self.data = data

        self.id = data["id"]

        self.body = data["body"]
        self.body_md = data["bodyMarkdown"]
        self._author = None
        self.is_internal = data["isInternal"]
        self.date = data["date"]

    async def author(self) -> 'Redditor':
        if self._author is None:
            if not self.data["author"]["isDeleted"]:
                self._author = self.conversation.reddit.redditor(
                    self.data["author"]["name"])
            else:
                return None
        return self._author



from typing import TYPE_CHECKING, AsyncIterator, Dict, List

from ..endpoints import API_PATH
from ..utils import snake_case_keys
from .apraw_base import aPRAWBase
from .redditor import Redditor
from .subreddit import Subreddit

if TYPE_CHECKING:
    from ..reddit import Reddit
    from .submission import Submission


class Comment(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict, submission: 'Submission' = None,
                 author: Redditor = None, subreddit: Subreddit = None, replies: List['Comment'] = None):
        super().__init__(reddit, data)

        self._submission = submission
        self._author = author
        self._subreddit = subreddit
        self._full_data = None
        self._replies = replies

        self.subreddit_name = data["subreddit"]
        self.url = "https://www.reddit.com" + data["permalink"]

    async def author(self) -> Redditor:
        if self._author is None:
            self._author = await self.reddit.redditor(self.data["author"])
        return self._author

    async def submission(self) -> 'Submission':
        if self._submission is None:
            link = await self.reddit.get_request(API_PATH["info"], id=self.data["link_id"])
            from .submission import Submission
            self._submission = Submission(
                self.reddit, link["data"]["children"][0]["data"])
        return self._submission

    async def subreddit(self) -> Subreddit:
        if self._subreddit is None:
            self._subreddit = await self.reddit.subreddit(self.subreddit_name)
        return self._subreddit

    async def full_data(self, refresh: bool = False) -> Dict:
        if self._full_data is None or refresh:
            self._full_data = await self.reddit.get_request(
                API_PATH["comment"].format(sub=self.data["subreddit"],
                                           submission=self.data["link_id"].replace(
                                               self.reddit.link_kind + "_", ""),
                                           id=self.data["id"]))
        return self._full_data

    async def refresh(self):
        await self.full_data(True)
        await self.replies(True)

    async def replies(self, refresh: bool = False) -> AsyncIterator['Comment']:
        if self._replies is None or refresh:
            fd = await self.full_data()

            def find(listing):
                data = listing["data"]

                for comment in data["children"]:
                    if comment["data"]["id"] == self.id:
                        return comment
                    else:
                        return find(comment["replies"])

            comment = find(fd[1])

            def get_replies(comment, replies=[]):
                if "kind" in comment["replies"] and comment["replies"]["kind"] == self.reddit.listing_kind:
                    for reply in comment["replies"]["data"]["children"]:
                        data = reply["data"]
                        replies.append(
                            Comment(self.reddit, data, replies=get_replies(data, [])))
                return replies

            self._replies = get_replies(comment["data"])

        for reply in self._replies:
            yield reply



import asyncio
import configparser
import logging
from datetime import datetime, timedelta
from typing import Any, Awaitable, Callable, Dict, List, Union

import aiohttp

from .endpoints import API_PATH, BASE_URL
from .models import (Comment, ListingGenerator, Redditor, Submission,
                     Subreddit, User)
from .utils import prepend_kind


class Reddit:

    def __init__(self, praw_key: str = "", username: str = "", password: str = "",
                 client_id: str = "", client_secret: str = "",
                 user_agent="aPRAW by Dan6erbond"):
        if praw_key != "":
            config = configparser.ConfigParser()
            config.read("praw.ini")

            self.user = User(self, config[praw_key]["username"], config[praw_key]["password"],
                             config[praw_key]["client_id"], config[praw_key]["client_secret"],
                             config[praw_key]["user_agent"] if "user_agent" in config[praw_key] else user_agent)
        else:
            self.user = User(self, username, password,
                             client_id, client_secret, user_agent)

        self.comment_kind = "t1"
        self.account_kind = "t2"
        self.link_kind = "t3"
        self.message_kind = "t4"
        self.subreddit_kind = "t5"
        self.award_kind = "t6"
        self.modaction_kind = "modaction"
        self.listing_kind = "Listing"
        self.wiki_revision_kind = "WikiRevision"

        self.subreddits = ListingGenerator(self, API_PATH["subreddits_new"])
        self.request_handler = RequestHandler(self.user)

    async def get_request(self, *args, **kwargs):
        return await self.request_handler.get_request(*args, **kwargs)

    async def post_request(self, *args, **kwargs):
        return await self.request_handler.post_request(*args, **kwargs)

    def get_listing_generator(self, *args, **kwargs):
        return ListingGenerator.get_listing_generator(self, *args, **kwargs)

    async def subreddit(self, display_name: str) -> Subreddit:
        resp = await self.get_request(API_PATH["subreddit_about"].format(sub=display_name))
        try:
            return Subreddit(self, resp["data"])
        except Exception as e:
            logging.error(e)
            return None

    async def info(self, id: str = "", ids: List[str] = [], url: str = ""):
        listing_generator = self.get_listing_generator(API_PATH["info"])

        if id:
            async for i in listing_generator(id=id):
                yield i
        elif ids:
            async for i in listing_generator(None, id=",".join(ids)):
                yield i
        elif url:
            async for i in listing_generator(url=url):
                yield i
        else:
            yield None

    async def submission(self, id: str = "", url: str = "") -> Submission:
        if id != "":
            async for link in self.info(prepend_kind(id, self.link_kind)):
                return link
        elif url != "":
            async for link in self.info(url=url):
                return link
        return None

    async def comment(self, id: str = "", url: str = "") -> Comment:
        if id != "":
            async for comment in self.info(prepend_kind(id, self.comment_kind)):
                return comment
        elif url != "":
            async for comment in self.info(url=url):
                return comment
        return None

    async def redditor(self, username: str) -> Redditor:
        resp = await self.get_request(API_PATH["user_about"].format(user=username))
        try:
            return Redditor(self, resp["data"])
        except Exception as e:
            # print("No Redditor data loaded for {}.".format(username))
            return None

    async def message(self, to: Union[str, Redditor], subject: str, text: str, from_sr: Union[str, Subreddit] = "") -> Dict:
        data = {
            "subject": subject,
            "text": text,
            "to": str(to)
        }
        if from_sr != "":
            data["from_sr"] = str(from_sr)
        resp = await self.post_request(API_PATH["compose"], data=data)
        return resp["success"]


class RequestHandler:

    def __init__(self, user):
        self.user = user
        self.queue = []

    async def get_request_headers(self) -> Dict:
        if self.user.token_expires <= datetime.now():
            url = "https://www.reddit.com/api/v1/access_token"
            session = self.user.get_auth_session()

            headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                "User-Agent": self.user.user_agent
            }

            resp = await session.post(url, data=self.user.password_grant, headers=headers)

            async with resp:
                if resp.status == 200:
                    self.user.access_data = await resp.json()
                    self.user.token_expires = datetime.now()
                    + timedelta(seconds=self.user.access_data["expires_in"])
                else:
                    raise Exception("Invalid user data.")

        return {
            "Authorization": "{} {}".format(self.user.access_data["token_type"], self.user.access_data["access_token"]),
            "User-Agent": self.user.user_agent
        }

    def update(self, data: Dict):
        self.user.ratelimit_remaining = int(
            float(data["x-ratelimit-remaining"]))
        self.user.ratelimit_used = int(data["x-ratelimit-used"])

        self.user.ratelimit_reset = datetime.now()
        + timedelta(seconds=int(data["x-ratelimit-reset"]))

    def check_ratelimit(
            func: Callable[[Any], Awaitable[Any]]) -> Callable[[Any], Awaitable[Any]]:
        async def execute_request(self, *args, **kwargs) -> Any:
            id = datetime.now().strftime('%Y%m%d%H%M%S')
            self.queue.append(id)

            if (self.user.ratelimit_remaining < 5):
                execution_time = self.user.ratelimit_reset
                + timedelta(seconds=len(self.queue))
                wait_time = (execution_time - datetime.now()).total_seconds()
                await asyncio.sleep(wait_time)

            result = await func(self, *args, **kwargs)
            self.queue.remove(id)
            return result

        return execute_request

    @check_ratelimit
    async def get_request(self, endpoint: str = "", **kwargs) -> Dict:
        kwargs["raw_json"] = 1
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        url = BASE_URL.format(endpoint, "&".join(params))

        headers = await self.get_request_headers()
        session = self.user.get_client_session()
        resp = await session.get(url, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()

    @check_ratelimit
    async def post_request(self, endpoint: str = "", url: str = "", data: Dict = {}, **kwargs) -> Dict:
        kwargs["raw_json"] = 1
        params = ["{}={}".format(k, kwargs[k]) for k in kwargs]

        if endpoint != "":
            url = BASE_URL.format(endpoint, "&".join(params))
        elif url != "":
            url = "{}?{}".format(url, "&".join(params))

        headers = await self.get_request_headers()
        session = self.user.get_client_session()
        resp = await session.post(url, data=data, headers=headers)

        async with resp:
            self.update(resp.headers)
            return await resp.json()



from typing import TYPE_CHECKING, Dict, List, Union

from ..endpoints import API_PATH
from ..utils import prepend_kind
from .apraw_base import aPRAWBase
from .redditor import Redditor

if TYPE_CHECKING:
    from ..reddit import Reddit
    from .subreddit import Subreddit


class SubredditWiki:

    def __init__(self, subreddit: 'Subreddit'):
        self.subreddit = subreddit

        self._data = None

        from .listing_generator import ListingGenerator
        self.revisions = ListingGenerator(
            subreddit.reddit, API_PATH["wiki_revisions"].format(
                sub=self.subreddit))

    async def data(self, refresh=False) -> Dict:
        if self._data is None:
            self._data = await self.subreddit.reddit.get_request(
                API_PATH["wiki"].format(sub=self.subreddit))
        return self._data

    async def __call__(self) -> List[str]:
        data = await self.data()
        return [page for page in data["data"]]

    async def page(self, page: str) -> 'SubredditWikiPage':
        resp = await self.subreddit.reddit.get_request(
            API_PATH["wiki_page"].format(sub=self.subreddit, page=page))
        return SubredditWikiPage(page, self.subreddit, resp["data"])

    async def create(self, page: str, content_md: str = "", reason: str = "") -> 'SubredditWikiPage':
        resp = await self.subreddit.reddit.post_request(
            API_PATH["wiki_edit"].format(sub=self.subreddit), data={
                "page": page,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else await self.page(page)


class SubredditWikiPage(aPRAWBase):

    def __init__(self, name: str, subreddit: 'Subreddit', data: Dict = None):
        super().__init__(subreddit.reddit, data)

        self.name = name
        self.subreddit = subreddit

        from .listing_generator import ListingGenerator
        self.revisions = ListingGenerator(
            subreddit.reddit, API_PATH["wiki_page_revisions"].format(
                sub=self.subreddit, page=self.name))

    async def _alloweditor(self, username: str, act: str):
        resp = await self.subreddit.reddit.post_request(
            API_PATH["wiki_alloweditor"].format(sub=self.subreddit, act=act), data={
                "page": self.name,
                "username": username
            })
        return True if not resp else resp

    async def add_editor(self, username: str):
        return await self._alloweditor(username, "add")

    async def del_editor(self, username: str):
        return await self._alloweditor(username, "del")

    async def edit(self, content_md: str = "", reason: str = "") -> bool:
        resp = await self.subreddit.reddit.post_request(
            API_PATH["wiki_edit"].format(sub=self.subreddit), data={
                "page": self.name,
                "content": content_md,
                "reason": reason
            })
        return resp if resp else True

    async def hide(self, revision: Union[str, 'WikipageRevision']):
        resp = await self.subreddit.reddit.post_request(
            API_PATH["wiki_hide"].format(sub=self.subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True

    async def revert(self, revision: Union[str, 'WikipageRevision']):
        resp = await self.subreddit.reddit.post_request(
            API_PATH["wiki_revert"].format(sub=self.subreddit), data={
                "page": self.name,
                "revision": str(revision)
            })
        return resp if resp else True


class WikipageRevision(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict = None):
        super().__init__(reddit, data)

        self.author = Redditor(reddit, data["author"]["data"])

    def __str__(self):
        return self.id



from typing import TYPE_CHECKING, AsyncIterator, Dict, List

from ..endpoints import API_PATH
from ..utils import snake_case_keys
from .apraw_base import aPRAWBase
from .comment import Comment
from .redditor import Redditor
from .subreddit import Subreddit

if TYPE_CHECKING:
    from ..reddit import Reddit


class Submission(aPRAWBase):
    def __init__(self, reddit: 'Reddit', data: Dict, full_data: Dict = None,
                 subreddit: Subreddit = None, author: Redditor = None):
        super().__init__(reddit, data)

        self._full_data = full_data
        self._comments = list()
        self._subreddit = subreddit
        self._author = author

        self.original_content = data["is_original_content"]

    async def full_data(self) -> Dict:
        if self._full_data is None:
            sub = await self.subreddit()
            self._full_data = await self.reddit.get_request(API_PATH["submission"].format(sub=sub.display_name, id=self.id))
        return self._full_data

    async def comments(self, reload=False, **kwargs) -> AsyncIterator[Comment]:
        if len(self._comments) <= 0 or reload:
            fd = await self.full_data()
            self._comments = list()

            for c in fd[1]["data"]["children"]:
                if c["kind"] == self.reddit.comment_kind:
                    self._comments.append(
                        Comment(
                            self.reddit,
                            c["data"],
                            submission=self))
                if c["kind"] == "more":
                    self._comments.extend(await self.morechildren(c["data"]["children"]))
        for c in self._comments:
            yield c

    async def morechildren(self, children) -> List[Comment]:
        comments = list()

        while len(children) > 0:
            cs = children[:100]
            children = children[100:]

            def get_comments(comment_list, comments):
                for i in comment_list:
                    if isinstance(i, list):
                        comments = get_comments(i, comments)
                    elif isinstance(i, dict) and "kind" in i and i["kind"] == self.reddit.comment_kind:
                        comments.append(
                            Comment(self.reddit, i["data"], submission=self))
                return comments

            data = await self.reddit.get_request(API_PATH["morechildren"], children=",".join(cs), link_id=self.name)
            comments = get_comments(data["jquery"], comments)

        return comments

    async def subreddit(self) -> Subreddit:
        if self._subreddit is None:
            self._subreddit = await self.reddit.subreddit(self.data["subreddit"])
        return self._subreddit

    async def author(self) -> Subreddit:
        if self._author is None:
            self._author = await self.reddit.redditor(self.data["author"])
        return self._author



from datetime import datetime
from typing import TYPE_CHECKING, Dict

import aiohttp

from ..endpoints import API_PATH
from .redditor import Redditor

if TYPE_CHECKING:
    from ..reddit import Reddit


class User:

    def __init__(self, reddit: 'Reddit', username: str, password: str, client_id: str,
                 client_secret: str, user_agent: str):
        self.reddit = reddit

        self.username = username
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret
        self.user_agent = user_agent

        if self.username == "" or self.password == "" or self.client_id == "" or self.client_secret == "":
            raise Exception(
                "No login information given or login information incomplete.")

        self.password_grant = "grant_type=password&username={}&password={}".format(
            self.username, self.password)

        self._auth_session = None
        self._client_session = None

        self._auth_user = None

        self.access_data = None
        self.token_expires = datetime.now()

        self.ratelimit_remaining = 0
        self.ratelimit_used = 0
        self.ratelimit_reset = datetime.now()

    def get_auth_session(self) -> aiohttp.ClientSession:
        if self._auth_session is None:
            auth = aiohttp.BasicAuth(
                login=self.client_id,
                password=self.client_secret)
            self._auth_session = aiohttp.ClientSession(auth=auth)
        return self._auth_session

    def get_client_session(self) -> aiohttp.ClientSession:
        if self._client_session is None:
            self._client_session = aiohttp.ClientSession()
        return self._client_session

    async def me(self) -> 'AuthenticatedUser':
        if not self._auth_user:
            data = await self.reddit.get_request(API_PATH["me"])
            self._auth_user = AuthenticatedUser(self.reddit, data)
        return self._auth_user


class AuthenticatedUser(Redditor):

    def __init__(self, reddit: 'Reddit', data: Dict):
        super().__init__(reddit, data)

        self._karma = None

    async def karma(self):
        if not self._karma:
            data = await self.reddit.get_request(API_PATH["me"])
            self._karma = AuthenticatedUser(self.reddit, data)
        return self._karma



import asyncio
from datetime import datetime
from typing import TYPE_CHECKING, AsyncIterator, Dict, List

from ..endpoints import API_PATH
from ..utils import snake_case_keys
from .apraw_base import aPRAWBase
from .modmail import SubredditModmail
from .redditor import Redditor
from .subreddit_wiki import SubredditWiki

if TYPE_CHECKING:
    from ..reddit import Reddit


class Subreddit(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict):
        super().__init__(reddit, data)

        self.quarantine = data["quarantine"] if "quarantine" in data else False

        self.mod = SubredditModeration(self)
        self.modmail = SubredditModmail(self)
        self.wiki = SubredditWiki(self)

        from .listing_generator import ListingGenerator
        self.comments = ListingGenerator(
            self.reddit, API_PATH["subreddit_comments"].format(
                sub=self.display_name), subreddit=self)
        self.new = ListingGenerator(self.reddit,
                                    API_PATH["subreddit_new"].format(sub=self.display_name), subreddit=self)
        self.hot = ListingGenerator(self.reddit,
                                    API_PATH["subreddit_hot"].format(sub=self.display_name), subreddit=self)
        self.rising = ListingGenerator(
            self.reddit, API_PATH["subreddit_rising"].format(sub=self.display_name))
        self.top = ListingGenerator(self.reddit,
                                    API_PATH["subreddit_top"].format(sub=self.display_name), subreddit=self)

    def __str__(self):
        return self.display_name

    async def moderators(self, **kwargs) -> AsyncIterator['SubredditModerator']:
        req = await self.reddit.get_request(API_PATH["subreddit_moderators"].format(sub=self.display_name), **kwargs)
        for u in req["data"]["children"]:
            yield SubredditModerator(self.reddit, u)

    async def message(self, subject, text, from_sr="") -> Dict:
        return await self.reddit.message(API_PATH["subreddit"].format(sub=self.display_name), subject, text, from_sr)


class SubredditModerator(aPRAWBase):

    def __init__(self, reddit: 'Reddit', data: Dict):
        super().__init__(reddit, data)

        self.added = data["date"]

    def __str__(self):
        return self.name

    async def redditor(self) -> Redditor:
        return await self.reddit.redditor(self.name)


class SubredditModeration:

    def __init__(self, subreddit):
        self.subreddit = subreddit

        from .listing_generator import ListingGenerator
        self.reports = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_reports"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)
        self.spam = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_spam"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)
        self.modqueue = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_modqueue"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)
        self.unmoderated = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_unmoderated"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)
        self.edited = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_edited"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)
        self.log = ListingGenerator(
            self.subreddit.reddit,
            API_PATH["subreddit_log"].format(
                sub=self.subreddit.display_name), subreddit=self.subreddit)


class ModAction:

    def __init__(self, data, subreddit=None):
        self.data = data
        self.subreddit = subreddit

        self.created_utc = datetime.utcfromtimestamp(data["created_utc"])

        d = snake_case_keys(data)
        for key in d:
            if not hasattr(self, key):
                setattr(self, key, d[key])

    async def mod(self) -> Redditor:
        return await self.subreddit.reddit.redditor(self.mod)



from typing import TYPE_CHECKING, Dict

from ..endpoints import API_PATH
from .apraw_base import aPRAWBase

if TYPE_CHECKING:
    from .subreddit import Subreddit
    from ..reddit import Reddit


class Redditor(aPRAWBase):
    def __init__(self, reddit: 'Reddit', data: Dict):
        super().__init__(reddit, data)

        self.reddit = reddit
        self.data = data

        self.is_suspended = "is_suspended" not in data or not data["is_suspended"]

        if "subreddit" in data and data["subreddit"]:
            sub = data["subreddit"]
            sub["id"] = sub["name"].replace("t5_", "")
            if "created_utc" not in sub:
                sub["created_utc"] = data["created_utc"]
            from .subreddit import Subreddit
            self.subreddit = Subreddit(self.reddit, sub)
        else:
            self.subreddit = None

        from .listing_generator import ListingGenerator
        self.comments = ListingGenerator(
            self.reddit,
            API_PATH["user_comments"].format(
                user=self))
        self.submissions = ListingGenerator(
            self.reddit,
            API_PATH["user_submissions"].format(
                user=self))

    def __str__(self):
        return self.name

    async def moderated_subreddits(self, **kwargs) -> 'Subreddit':
        req = await self.reddit.get_request(API_PATH["moderated"].format(user=self), **kwargs)
        for s in req["data"]:
            yield await self.reddit.subreddit(s["sr"])

    async def message(self, subject, text, from_sr="") -> Dict:
        return await self.reddit.message(self.name, subject, text, from_sr)



import asyncio
from typing import TYPE_CHECKING, Any, AsyncIterator, Callable, List, Union

from ..utils import prepend_kind
from .apraw_base import aPRAWBase
from .comment import Comment
from .submission import Submission
from .subreddit import ModAction, Subreddit
from .subreddit_wiki import WikipageRevision

if TYPE_CHECKING:
    from ..reddit import Reddit


class ListingGenerator:

    def __init__(self, reddit: 'Reddit', endpoint: str,
                 max_wait: int = 16, kind_filter: List[str] = [],
                 subreddit=None):
        self.reddit = reddit
        self.endpoint = endpoint
        self.max_wait = max_wait
        self.kind_filter = kind_filter
        self.subreddit = subreddit

    @classmethod
    def get_listing_generator(cls, reddit: 'Reddit', endpoint: str,
                              max_wait: int = 16, kind_filter: List[str] = [],
                              subreddit=None) -> Callable[[Any], AsyncIterator[aPRAWBase]]:
        async def get_listing(limit: int = 25, **kwargs) -> AsyncIterator[aPRAWBase]:
            last = None
            while True:
                kwargs["limit"] = limit if limit is not None else 100
                if last is not None:
                    kwargs["after"] = last
                req = await reddit.get_request(endpoint, **kwargs)
                if len(req["data"]["children"]) <= 0:
                    break
                for i in req["data"]["children"]:
                    wiki_page = "page" in i

                    if wiki_page and kind_filter and reddit.wiki_revision_kind not in kind_filter:
                        continue
                    elif not wiki_page and kind_filter and i["kind"] in kind_filter:
                        continue

                    if wiki_page:
                        last = prepend_kind(i["id"], reddit.wiki_revision_kind)
                    elif i["kind"] in [reddit.link_kind,
                                       reddit.subreddit_kind, reddit.comment_kind]:
                        last = i["data"]["name"]
                    elif i["kind"] == reddit.modaction_kind:
                        last = i["data"]["id"]

                    if limit is not None:
                        limit -= 1

                    if wiki_page:
                        yield WikipageRevision(reddit, i)
                    elif i["kind"] == reddit.link_kind:
                        yield Submission(reddit, i["data"], subreddit=subreddit)
                    elif i["kind"] == reddit.subreddit_kind:
                        yield Subreddit(reddit, i["data"])
                    elif i["kind"] == reddit.comment_kind:
                        yield Comment(reddit, i["data"], subreddit=subreddit)
                    elif i["kind"] == reddit.modaction_kind:
                        yield ModAction(i["data"], subreddit)
                    else:
                        yield aPRAWBase(i["data"] if "data" in i else i)
                if limit is not None and limit < 1:
                    break

        return get_listing

    async def get(self, *args, **kwargs) -> AsyncIterator[aPRAWBase]:
        async for i in ListingGenerator.get_listing_generator(**vars(self))(*args, **kwargs):
            yield i

    __call__ = get

    async def stream(self, **kwargs) -> AsyncIterator[aPRAWBase]:
        wait = 0
        ids = list()

        while True:
            found = False
            async for s in self.get(100, **kwargs):
                if s.id in ids:
                    break
                if len(ids) >= 301:
                    ids = ids[1:]
                ids.append(s.id)
                found = True
                yield s

            if found:
                wait = 1
            else:
                wait *= 2
                if wait > self.max_wait:
                    wait = 1

            await asyncio.sleep(wait)
